#include <Quickdraw.h>
#include <Menus.h>
#include <Windows.h>
#include <ToolUtils.h>
#include <Desk.h>
#include <SegLoad.h>
#include <OSUtils.h>
#include <Memory.h>
#include <Fonts.h>
#include <Events.h>
#include <OSEvents.h>
#include <TextEdit.h>
#include <Dialogs.h>

void AdjustMenus(void); 
void HandleMenu (long mSelect);
void SetUpMenus(void);
void InitMacintosh(void);
void HandleMouseDown (EventRecord *theEvent);
void HandleEvent(void);
void DrawBullseye (short active);
void SetUpWindow(void);

void
main( void)

{
	InitMacintosh();
	SetUpMenus();
	SetUpWindow();
	
	for (;;)
		HandleEvent();
}
/* end main */

extern	WindowPtr bullseyeWindow;
extern	int	width;


MenuHandle	appleMenu, fileMenu, editMenu, widthMenu;

enum	{
	appleID = 1,
	fileID,
	editID,
	widthID
	};

enum	{
	openItem = 1,
	closeItem,
	quitItem = 4
	};


/****
 * SetUpMenus()
 *
 *	Set up the menus. Normally, wed use a resource file, but
 *	for this example well supply hardwired strings.
 *
 ****/

void SetUpMenus(void)

{
	InsertMenu(appleMenu = NewMenu(appleID, "\p\024"), 0);
	InsertMenu(fileMenu = NewMenu(fileID, "\pFile"), 0);
	InsertMenu(editMenu = NewMenu(editID, "\pEdit"), 0);
	InsertMenu(widthMenu = NewMenu(widthID, "\pWidth"), 0);
	DrawMenuBar();
	AddResMenu(appleMenu, 'DRVR');
	AppendMenu(fileMenu, "\pOpen/O;Close/W;(-;Quit/Q");
	AppendMenu(editMenu, "\pUndo/Z;(-;Cut/X;Copy/C;Paste/V;Clear");
	AppendMenu(widthMenu, "\p1/1;2/2;3/3;4/4;5/5;6/6;7/7;8/8;9/9;10/0;11;12");
}
/* end SetUpMenus */


/****
 *  AdjustMenus()
 *
 *	Enable or disable the items in the Edit menu if a DA window
 *	comes up or goes away. Our application doesn't do anything with
 *	the Edit menu.
 *
 ****/

static void enable (MenuHandle menu, short item, short ok);

void AdjustMenus(void)
{
	register WindowPeek wp = (WindowPeek) FrontWindow();
	short kind = wp ? wp->windowKind : 0;
	Boolean DA = kind < 0;
	
	enable(editMenu, 1, DA);
	enable(editMenu, 3, DA);
	enable(editMenu, 4, DA);
	enable(editMenu, 5, DA);
	enable(editMenu, 6, DA);
	
	enable(fileMenu, openItem, !((WindowPeek) bullseyeWindow)->visible);
	enable(fileMenu, closeItem, DA || ((WindowPeek) bullseyeWindow)->visible);

	CheckItem(widthMenu, width, true);
}


static
void enable(MenuHandle menu, short item, short ok)
{
	if (ok)
		EnableItem(menu, item);
	else
		DisableItem(menu, item);
}


/*****
 * HandleMenu(mSelect)
 *
 *	Handle the menu selection. mSelect is what MenuSelect() and
 *	MenuKey() return: the high word is the menu ID, the low word
 *	is the menu item
 *
 *****/

void HandleMenu (long mSelect)

{
	int			menuID = HiWord(mSelect);
	int			menuItem = LoWord(mSelect);
	Str255		name;
	GrafPtr		savePort;
	WindowPeek	frontWindow;
	
	switch (menuID)
	  {
	  case	appleID:
		GetPort(&savePort);
		GetItem(appleMenu, menuItem, name);
		OpenDeskAcc(name);
		SetPort(savePort);
		break;
	
	  case	fileID:
		switch (menuItem)
		  {
		  case	openItem:
			ShowWindow(bullseyeWindow);
			SelectWindow(bullseyeWindow);
			break;
  				  			
		  case	closeItem:
			if ((frontWindow = (WindowPeek) FrontWindow()) == 0L)
			  break;
			  
			if (frontWindow->windowKind < 0)
			  CloseDeskAcc(frontWindow->windowKind);
			else if ((frontWindow = (WindowPeek) bullseyeWindow) != NULL)
			  HideWindow(bullseyeWindow);
  				  	break;
  				  	
		  case	quitItem:
			ExitToShell();
			break;
		  }
		break;
  				
	  case	editID:
		if (!SystemEdit(menuItem-1))
		  SysBeep(5);
		break;
		
	  case	widthID:
		CheckItem(widthMenu, width, false);
		width = menuItem;
		InvalRect(&bullseyeWindow->portRect);
		break;
	  }
}
/* end HandleMenu */

/*****
 * bullWindow.c
 *
 *		The window routines for the Bullseye demo
 *
 *****/

/*****
 * bullWindow.h
 *
 *		Public interfaces for bullWindow.c
 *
 *****/

WindowPtr	bullseyeWindow;
Rect		dragRect;
Rect		windowBounds = { 40, 40, 150, 150 };
Rect		circleStart = {10, 10, 100, 100};
int			width = 10;

/****
 * SetUpWindow()
 *
 *	Create the Bullseye window, and open it.
 *
 ****/

void SetUpWindow(void)

{
	dragRect = qd.screenBits.bounds;
	
	bullseyeWindow = NewWindow(0L, &windowBounds, "\pBullseye", true, noGrowDocProc, (WindowPtr) -1L, true, 0);
	SetPort(bullseyeWindow);
}
/* end SetUpWindow */


/*****
 * DrawBullseye()
 *
 *	Draws the bullseye.
 *
 *****/

void DrawBullseye(short active)

{
	Rect	myRect;
	int		color = true;
	
	SetPort(bullseyeWindow);
	EraseRect(&circleStart);
	myRect = circleStart;
	
	while(myRect.left < myRect.right)
	  {
	  ConstPatternParam pat = color ? (active ? &qd.black : &qd.gray) : &qd.white;
	  FillOval(&myRect, pat);
	  InsetRect(&myRect, width, width);
	  color = !color;
	  } 
}
/* end DrawBullseye */
/*****
 * bullseye.c
 *
 *	A simple demonstration program to play with the debugger
 *
 *
 *****/

extern	WindowPtr	bullseyeWindow;
extern	Rect		dragRect;

/****
 * InitMacintosh()
 *
 * Initialize all the managers & memory
 *
 ****/

void InitMacintosh(void)

{
	MaxApplZone();
	
	InitGraf(&qd.thePort);
	InitFonts();
	FlushEvents(everyEvent, 0);
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(0L);
	InitCursor();

}
/* end InitMacintosh */


/****
 * HandleMouseDown (theEvent)
 *
 *	Take care of mouseDown events.
 *
 ****/

void HandleMouseDown (EventRecord	*theEvent)

{
	WindowPtr	theWindow;
	int			windowCode = FindWindow (theEvent->where, &theWindow);
	
    switch (windowCode)
      {
	  case inSysWindow: 
	    SystemClick (theEvent, theWindow);
	    break;
	    
	  case inMenuBar:
	  	AdjustMenus();
	    HandleMenu(MenuSelect(theEvent->where));
	    break;
	    
	  case inDrag:
	  	if (theWindow == bullseyeWindow)
	  	  DragWindow(bullseyeWindow, theEvent->where, &dragRect);
	  	  break;
	  	  
	  case inContent:
	  	if (theWindow == bullseyeWindow)
	  	  {
	  	  if (theWindow != FrontWindow())
	  	    SelectWindow(bullseyeWindow);
	  	  else
	  	    InvalRect(&bullseyeWindow->portRect);
	  	  }
	  	break;
	  	
	  case inGoAway:
	  	if (theWindow == bullseyeWindow && 
	  		TrackGoAway(bullseyeWindow, theEvent->where))
		  HideWindow(bullseyeWindow);
	  	  break;
      }
}
/* end HandleMouseDown */


/****
 * HandleEvent()
 *
 *		The main event dispatcher. This routine should be called
 *		repeatedly (it  handles only one event).
 *
 *****/

void HandleEvent(void)

{
	int			ok;
	EventRecord	theEvent;

	HiliteMenu(0);
	SystemTask ();		/* Handle desk accessories */
	
	ok = GetNextEvent (everyEvent, &theEvent);
	if (ok)
	  switch (theEvent.what)
	    {
		case mouseDown:
			HandleMouseDown(&theEvent);
			break;
			
		case keyDown: 
		case autoKey:
		    if ((theEvent.modifiers & cmdKey) != 0)
		      {
		      AdjustMenus();
			  HandleMenu(MenuKey((char) (theEvent.message & charCodeMask)));
			  }
			break;
			
		case updateEvt:
			BeginUpdate(bullseyeWindow);
			DrawBullseye(((WindowPeek) bullseyeWindow)->hilited);
			EndUpdate(bullseyeWindow);
		    break;
		    
		case activateEvt:
			InvalRect(&bullseyeWindow->portRect);
			break;
	    }
}
/* end HandleEvent */

