/* Definitions of target machine for GNU compiler.  For Macintosh MPW,
   extensively modified from the generic 68k tm.h.
   Copyright (C) 1987, 1988 Free Software Foundation, Inc.
   Copyright (C) 1989, 1990 Apple Computer, Inc.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* This comment is here to see if it will keep Sun's cpp from dying.  */

/* These actually only affect the "gcc" program in Unix, are unused in
   the MPW world. */

typedef enum { noreg, pd0, pd1, pd2, pa0, pa1, k=10000} pregs;

#define MAXPARAMS 5

typedef struct parameter {
	struct parameter *last;
	pregs	funcreturn;
	pregs	params[MAXPARAMS];
	char	name[1];
} parameter;

parameter *
lookupparameter( const char *name );
void
printreg( pregs r );

#define CPP_SPEC "%{m68881:-Dmc68881}%{m68020:-Dmc68020}"

#define CC1_SPEC "%{s*}"

/* This is true when code gen can treat symbol/label refs as constants. */

#define SYMBOLS_ARE_CONSTANTS 1

/* Define this when labels/symbols have to be explicitly imported into
   a module. */

#undef IMPORT_NAMES

/* Define this when code labels and data labels need to be treated
   differently in some way. */

#define MARK_LABEL_SECTION

/* Names to predefine in the preprocessor for this target machine.  */
/* Note that this is exactly MPW's set, plus something to distinguish */
/* from "regular" MPW. */

#define CPP_PREDEFINES \
  "-DMC68000 -Dmc68000 -Dm68k -Dmacintosh -Dapplec -Dmpwgcc"

/* So *library call generation* uses memcpy instead of bcopy, etc. */

#define TARGET_MEM_FUNCTIONS

/* Print subsidiary information on the compiler version in use.  */
#define TARGET_VERSION fprintf (stderr, " (680x0, MPW Asm syntax)");

/* Declare Mac-specific globals used in the compiler. */
extern char *segment_name;
extern int generate_trace_calls;
extern int ignore_trace_pragmas;
extern int long_double_type_size;
extern int current_static_decl;

/* Should override defn in c-decl.c */

#define LONG_DOUBLE_TYPE_SIZE (long_double_type_size)

/* Run-time compilation parameters selecting different hardware subsets.  */

/* MPW compatibility means plain 68000 + SANE + Macsbug symbols. */

#define TARGET_DEFAULT 020

extern int target_flags;

/* Macros used in the machine description to test the flags.  */

/* Compile for a 68020 (not a 68000 or 68010).   */
#define TARGET_68020 (target_flags & 1)
/* Compile 68881 insns for floating point (not library calls).  */
#define TARGET_68881 (target_flags & 2)
/* Compile using 68020 bitfield insns.  */
#define TARGET_BITFIELD (target_flags & 4)
/* Compile with 16-bit `int'.  */
#define TARGET_SHORT (target_flags & 010)
/* Include MacsBuggable symbols. */
#define TARGET_MACSBUG (target_flags & 020)
/* Compile transcendental fns directly to 881 stuff. */
#define TARGET_ELEMS881 (target_flags & 040)
/* Produce 32-bit references to global data. */
#define TARGET_32BITDATA (target_flags & 0100)
/* Generate code for SANE calls directly. */
#define TARGET_SANE (target_flags & 0200)
/* Generate code for integer library calls directly. */
#define TARGET_INTLIB (target_flags & 0400)
/* Generate pc-relative refs for strings after fn name. */
#define TARGET_B (target_flags & 01000)
/* Generate obscure addressing modifications. */
#define TARGET_FX30 (target_flags & 02000)
/* Suppress all segment directives in the output. */
#define TARGET_NOSEG (target_flags & 04000)
/* gas compatible output */
#define TARGET_GAS (target_flags & 010000)
/* long double conventions */
#define TARGET_LONG_DOUBLE (target_flags & 020000)
/* no case tables */
#define TARGET_NODISPATCH (target_flags & 040000)

/* Macro to define tables used to set the flags.
   This is a list in braces of pairs in braces,
   each pair being { "NAME", VALUE }
   where VALUE is the bits to set or minus the bits to clear.
   An empty string NAME is used to identify the default VALUE.  */

#define TARGET_SWITCHES  \
  { { "68020", 5},				\
    { "68000", -5},				\
    { "68881", 2},				\
    { "sane", 0200 },  \
    { "insane", -0200 },  \
    { "bitfield", 4},				\
    { "nobitfield", -4},			\
    { "short", 010},				\
    { "noshort", -010},				\
    { "bgfull", 020 }, \
    { "bgoff", -020 }, \
    { "elems881", 040},  \
    { "m", 0100},  \
    { "intlib", 0400 },  \
    { "nointlib", -0400 },  \
    { "b", 01000 },  \
    { "fx30", 02000 },  \
    { "noseg", 04000 },  \
    { "gas", 014000 },  \
    { "nogas", -010000 },  \
    { "long-double", 020000 },  \
    { "nolong-double", -020000 },  \
    { "nodispatch", 040000 },  \
    { "", TARGET_DEFAULT}}

/* In the MPW world, "long double" size depends on compiler flag (yech). */
/* elems881 implies m68881. */
/* MPW Asm seems to require -m68020 when doing large memory model (-m). */
/* One or the other of sane/mc68881 must be on, but not both. */

#define OVERRIDE_OPTIONS		\
{					\
  if (TARGET_ELEMS881) target_flags |= 2;  \
  long_double_type_size = (TARGET_68881 ? 96 : 80);  \
  mode_size[(int) XFmode] = (TARGET_68881 ? 12 : 10);  \
  mode_unit_size[(int) XFmode] = (TARGET_68881 ? 12 : 10);  \
  if (TARGET_68881) target_flags &= ~0200;  \
  if (TARGET_SANE) target_flags &= ~2;  \
  if (TARGET_32BITDATA) target_flags |= 5;  \
  if (TARGET_GAS) target_flags |= 04000;  \
}

/* target machine storage layout */

/* Define this if most significant bit is lowest numbered
   in instructions that operate on numbered bit-fields.
   This is true for 68020 insns such as bfins and bfexts.
   We make it true always by avoiding using the single-bit insns
   except in special cases with constant bit numbers.  */
#define BITS_BIG_ENDIAN

/* Define this if most significant byte of a word is the lowest numbered.  */
/* That is true on the 68000.  */
#define BYTES_BIG_ENDIAN

/* Define this if most significant word of a multiword number is numbered.  */
/* For 68000 we can decide arbitrarily
   since there are no machine instructions for them.  */
/* #define WORDS_BIG_ENDIAN */

/* number of bits in an addressible storage unit */
#define BITS_PER_UNIT 8

/* Width in bits of a "word", which is the contents of a machine register.
   Note that this is not necessarily the width of data type `int';
   if using 16-bit ints on a 68000, this would still be 32.
   But on a machine with 16-bit registers, this would be 16.  */
#define BITS_PER_WORD 32

/* Width of a word, in units (bytes).  */
#define UNITS_PER_WORD 4

/* Width in bits of a pointer.
   See also the macro `Pmode' defined below.  */
#define POINTER_SIZE 32

/* Allocation boundary (in *bits*) for storing pointers in memory.  */
#define POINTER_BOUNDARY 16

/* Allocation boundary (in *bits*) for storing arguments in argument list.  */
/* Note that prototype promotion will do traditional alignment if necessary,
   but pascal fns will still need padding. */
#define PARM_BOUNDARY 16

/* Boundary (in *bits*) on which stack pointer should be aligned.  */
#define STACK_BOUNDARY 16

/* Allocation boundary (in *bits*) for the code of a function.  */
#define FUNCTION_BOUNDARY 16

/* Alignment of field after `int : 0' in a structure.  */
#define EMPTY_FIELD_BOUNDARY 16

/* Every structure or union's size must be a multiple of 2 bytes.  */
#define STRUCTURE_SIZE_BOUNDARY 16

/* No data type wants to be aligned rounder than this.  */
#define BIGGEST_ALIGNMENT 16

/* Define this if move instructions will actually fail to work
   when given unaligned data.  */
#define STRICT_ALIGNMENT

/* Define number of bits in most basic integer type.
   (If undefined, default is BITS_PER_WORD).  */

#define INT_TYPE_SIZE (TARGET_SHORT ? 16 : 32)

/* Define the largest mode that can be treated as a regular integer. */

/* MPW C tends to think of larger things as block mode objects, so we
   should do this to maintain cross-callability. */
#define MAX_FIXED_MODE_SIZE (GET_MODE_BITSIZE (SImode))

/* Standard register usage.  */

/* Number of actual hardware registers.
   The hardware registers are assigned numbers for the compiler
   from 0 to just below FIRST_PSEUDO_REGISTER.
   All registers that the compiler knows about must be given numbers,
   even those that are not normally considered general registers.
   For the 68000, we give the data registers numbers 0-7,
   the address registers numbers 010-017,
   and the 68881 floating point registers numbers 020-027.  */
#define FIRST_PSEUDO_REGISTER 24

/* 1 for registers that have pervasive standard uses
   and are not available for the register allocator.
   On the 68000, only the stack pointer is such, but A5 is special to Macs. */
#define FIXED_REGISTERS  \
 {0, 0, 0, 0, 0, 0, 0, 0, \
  0, 0, 0, 0, 0, 1, 1, 1, \
  0, 0, 0, 0, 0, 0, 0, 0, }

/* 1 for registers not available across function calls.
   These must include the FIXED_REGISTERS and also any
   registers that can be used without being saved.
   The latter must include the registers where values are returned
   and the register where structure-value addresses are passed.
   Aside from that, you can include as many other registers as you like.  */
/* This includes all of MPW's scratch registers. */
#define CALL_USED_REGISTERS \
 {1, 1, 1, 0, 0, 0, 0, 0, \
  1, 1, 0, 0, 0, 1, 1, 1, \
  1, 1, 1, 1, 0, 0, 0, 0, }

/* Make sure everything's fine if we *don't* have a given processor.
   This assumes that putting a register in fixed_regs will keep the
   compilers mitt's completely off it.  We don't bother to zero it out
   of register classes.  If TARGET_68881 is not set,
   the compiler won't touch since no instructions that use these
   registers will be valid.  */
#define CONDITIONAL_REGISTER_USAGE \
{ 						\
  int i; 					\
  HARD_REG_SET x; 				\
    { 						\
      COPY_HARD_REG_SET (x, reg_class_contents[(int)NO_REGS]); \
      for (i = 0; i < FIRST_PSEUDO_REGISTER; i++ ) \
       if (TEST_HARD_REG_BIT (x, i)) 		\
	fixed_regs[i] = call_used_regs[i] = 1; 	\
    } 						\
}

/* Return number of consecutive hard regs needed starting at reg REGNO
   to hold something of mode MODE.
   This is ordinarily the length in words of a value of mode MODE
   but can be less for certain modes in special long registers.

   On the 68000, ordinary registers hold 32 bits worth;
   for the 68881 registers, a single register is always enough for
   anything that can be stored in them at all.  */
#define HARD_REGNO_NREGS(REGNO, MODE)   \
  ((REGNO) >= 16 ? 1				\
   : ((GET_MODE_SIZE (MODE) + UNITS_PER_WORD - 1) / UNITS_PER_WORD))

/* Value is 1 if hard register REGNO can hold a value of machine-mode MODE.
   On the 68000, the cpu registers can hold any mode but the 68881 registers
   can hold only float modes.  And the 68881 registers can't hold anything
   if 68881 use is disabled.  Actually, we have to exclude d6 and/or d7 if
   we're using 2 or 3 registers to hold a value, since the "next" register(s)
   would be a0/a1, and things would get really confusing. */
#define HARD_REGNO_MODE_OK(REGNO, MODE) \
  (((REGNO) < 16 &&  \
    ! (GET_MODE_SIZE (MODE) > 4 && (REGNO) >= 6) &&  \
    ! (GET_MODE_SIZE (MODE) > 8 && (REGNO) >= 6))  \
   || (TARGET_68881 && GET_MODE_CLASS (MODE) == MODE_FLOAT))

/* Value is 1 if it is a good idea to tie two pseudo registers
   when one has mode MODE1 and one has mode MODE2.
   If HARD_REGNO_MODE_OK could produce different values for MODE1 and MODE2,
   for any hard reg, then this must be 0 for correct output.

   Going by the above defn, we can tie modes for general regs (what about
   d6/d7 tho?) and equal modes for float regs. */
#define MODES_TIEABLE_P(MODE1, MODE2)			\
   (TARGET_68881	?				\
   ((GET_MODE_CLASS (MODE1) == MODE_FLOAT)           \
       == (GET_MODE_CLASS (MODE2) == MODE_FLOAT)) :  \
   ((GET_MODE_SIZE (MODE1) < 8) && (GET_MODE_SIZE (MODE2) < 8)))

/* Specify the registers used for certain standard purposes.
   The values of these macros are register numbers.  */

/* m68000 pc isn't overloaded on a register.  */
/* #define PC_REGNUM  */

/* Register to use for pushing function arguments.  */
#define STACK_POINTER_REGNUM 15

/* Base register for access to local variables of the function.  */
#define FRAME_POINTER_REGNUM 14

/* Value should be nonzero if functions must have frame pointers.
   Zero means the frame pointer need not be set up (and parms
   may be accessed via the stack pointer) in functions that seem suitable.
   This is computed in `reload', in reload1.c.  */
#define FRAME_POINTER_REQUIRED 0

/* Base register for access to arguments of the function.  */
#define ARG_POINTER_REGNUM 14

/* Register in which static-chain is passed to a function.  */
#define STATIC_CHAIN_REGNUM 8

/* Register in which address to store a structure value
   arrives in the function.

   The MPW 3.0 calling convention is to have things on the stack top.  */
#define STRUCT_VALUE_INCOMING	\
   gen_rtx (MEM, Pmode, gen_rtx (PLUS, Pmode, frame_pointer_rtx,	\
				 gen_rtx (CONST_INT, VOIDmode, 8)))

/* Place in which caller passes the structure value address.
   Actually, all that matters about this value is it its rtx_code:
   MEM means push the value on the stack like an argument.  */
#define STRUCT_VALUE  \
   gen_rtx (MEM, Pmode, gen_rtx (PRE_DEC, Pmode, stack_pointer_rtx))

/* To be MPW-compatible, multi-word structures cannot live in regs. */

#define RETURN_IN_MEMORY(TYPE) \
  ((TREE_CODE (TYPE) == RECORD_TYPE || TREE_CODE (TYPE) == UNION_TYPE) && \
   (GET_MODE_SIZE (TYPE_MODE (TYPE)) > UNITS_PER_WORD))

/* Pad shorts and chars upwards, but nobody else. */

#define FUNCTION_ARG_PADDING(MODE,SIZE)  \
    (((MODE) == BLKmode							\
    ? (GET_CODE (SIZE) == CONST_INT					\
       && INTVAL (SIZE) < PARM_BOUNDARY / BITS_PER_UNIT)		\
    : GET_MODE_BITSIZE (MODE) < PARM_BOUNDARY)				\
   ? upward : none)


/* Define the classes of registers for register constraints in the
   machine description.  Also define ranges of constants.

   One of the classes must always be named ALL_REGS and include all hard regs.
   If there is more than one class, another class must be named NO_REGS
   and contain no registers.

   The name GENERAL_REGS must be the name of a class (or an alias for
   another name such as ALL_REGS).  This is the class of registers
   that is allowed by "g" or "r" in a register constraint.
   Also, registers outside this class are allocated only when
   instructions express preferences for them.

   The classes must be numbered in nondecreasing order; that is,
   a larger-numbered class must never be contained completely
   in a smaller-numbered class.

   For any two classes, it is very desirable that there be another
   class that represents their union.  */

/* The 68000 has three kinds of registers, so eight classes would be
   a complete set.  One of them (addr_or_fp) is not needed.  */

enum reg_class {
  NO_REGS, FP_REGS,
  DATA_REGS, DATA_OR_FP_REGS,
  ADDR_REGS, GENERAL_REGS,
  ALL_REGS,
  LIM_REG_CLASSES
#ifdef MPW_C31
  , reg_class_intifier = 1000000
#endif /* MPW_C31 */
};

#define N_REG_CLASSES (int) LIM_REG_CLASSES

/* Give names of register classes as strings for dump file.   */

#define REG_CLASS_NAMES \
 { "NO_REGS", "FP_REGS",  \
   "DATA_REGS", "DATA_OR_FP_REGS",  \
   "ADDR_REGS", "GENERAL_REGS",  \
   "ALL_REGS" }

/* Define which registers fit in which classes.
   This is an initializer for a vector of HARD_REG_SET
   of length N_REG_CLASSES.  */

#define REG_CLASS_CONTENTS \
{					\
 0,		/* NO_REGS */		\
 0xff0000,	/* FP_REGS */		\
 0x0000ff,	/* DATA_REGS */		\
 0xff00ff,	/* DATA_OR_FP_REGS */	\
 0x00ff00,	/* ADDR_REGS */		\
 0x00ffff,	/* GENERAL_REGS */	\
 0xffffff,	/* ALL_REGS */		\
}

/* The same information, inverted:
   Return the class number of the smallest class containing
   reg number REGNO.  This could be a conditional expression
   or could index an array.  */

extern enum reg_class regno_reg_class[];
#define REGNO_REG_CLASS(REGNO) (regno_reg_class[(REGNO)>>3])

/* The class value for index registers, and the one for base regs.  */

#define INDEX_REG_CLASS GENERAL_REGS
#define BASE_REG_CLASS ADDR_REGS

/* Get reg_class from a letter such as appears in the machine description.
   We do a trick here to modify the effective constraints on the
   machine description; we zorch the constraint letters that aren't
   appropriate for a specific target.  This allows us to guarrantee
   that a specific kind of register will not be used for a given taget
   without fiddling with the register classes above. */

#define REG_CLASS_FROM_LETTER(C) \
  ((C) == 'a' ? ADDR_REGS :			\
   ((C) == 'd' ? DATA_REGS :			\
    ((C) == 'f' ? (TARGET_68881 ? FP_REGS : NO_REGS) :  \
     NO_REGS)))

/* The letters I, J, K, L and M in a register constraint string
   can be used to stand for particular ranges of immediate operands.
   This macro defines what the ranges are.
   C is the letter, and VALUE is a constant value.
   Return 1 if VALUE is in the range specified by C.

   For the 68000, `I' is used for the range 1 to 8
   allowed as immediate shift counts and in addq.
   `J' is used for the range of signed numbers that fit in 16 bits.
   `K' is for numbers that moveq can't handle.
   `L' is for range -8 to -1, range of values that can be added with subq.  */

#define CONST_OK_FOR_LETTER_P(VALUE, C)  \
  ((C) == 'I' ? (VALUE) > 0 && (VALUE) <= 8 :    \
   (C) == 'J' ? (VALUE) >= -0x8000 && (VALUE) <= 0x7FFF :	\
   (C) == 'K' ? (VALUE) < -0x80 || (VALUE) >= 0x80 :	\
   (C) == 'L' ? (VALUE) < 0 && (VALUE) >= -8 :  \
   (C) == 'M' ? 1 : 0)

/*
 * A small bit of explanation:
 * "G" defines all of the floating constants that are *NOT* 68881
 * constants.  this is so 68881 constants get reloaded and the
 * fpmovecr is used.
 */
#define CONST_DOUBLE_OK_FOR_LETTER_P(VALUE, C)  \
  ((C) == 'G' ? ! (TARGET_68881 && standard_68881_constant_p (VALUE)) : 0)

/* Given an rtx X being reloaded into a reg required to be
   in class CLASS, return the class of reg to actually use.
   In general this is just CLASS; but on some machines
   in some cases it is preferable to use a more restrictive class.
   On the 68000 series, use a data reg if possible when the
   value is a constant in the range where moveq could be used
   and we ensure that QImodes are reloaded into data regs.  */
#define PREFERRED_RELOAD_CLASS(X,CLASS)  \
  ((GET_CODE (X) == CONST_INT			\
    && (unsigned) (INTVAL (X) + 0x80) < 0x100	\
    && (CLASS) != ADDR_REGS)			\
   ? DATA_REGS					\
   : GET_MODE (X) == QImode			\
   ? DATA_REGS					\
   : (CLASS))

/* Return the maximum number of consecutive registers
   needed to represent mode MODE in a register of class CLASS.  */
/* On the 68000, this is the size of MODE in words,
   except in the FP regs, where a single reg is always enough.  */
#define CLASS_MAX_NREGS(CLASS, MODE)	\
 ((CLASS) == FP_REGS ? 1 \
  : ((GET_MODE_SIZE (MODE) + UNITS_PER_WORD - 1) / UNITS_PER_WORD))

/* Stack layout; function entry, exit and calling.  */

/* Define this if pushing a word on the stack
   makes the stack pointer a smaller address.  */
#define STACK_GROWS_DOWNWARD

/* Define this if the nominal address of the stack frame
   is at the high-address end of the local variables;
   that is, each additional local variable allocated
   goes at a more negative offset in the frame.  */
#define FRAME_GROWS_DOWNWARD

/* Offset within stack frame to start allocating local variables at.
   If FRAME_GROWS_DOWNWARD, this is the offset to the END of the
   first local allocated.  Otherwise, it is the offset to the BEGINNING
   of the first local allocated.  */
#define STARTING_FRAME_OFFSET 0

/* If we generate an insn to push BYTES bytes,
   this says how many the stack pointer really advances by.
   On the 68000, sp@- in a byte insn really pushes a word.  */
#define PUSH_ROUNDING(BYTES) (((BYTES) + 1) & ~1)

/* Offset of first parameter from the argument pointer register value.  */
#define FIRST_PARM_OFFSET(FNDECL) 8

/* Value is 1 if returning from a function call automatically
   pops the arguments described by the number-of-args field in the call.
   FUNTYPE is the data type of the function (as a tree),
   or for a library call it is an identifier node for the subroutine name.

   Always false on the Mac, even for pascal fns (really?). */

#define RETURN_POPS_ARGS(FUNTYPE) 0

/* Define how to find the value returned by a function.
   VALTYPE is the data type of the value (as a tree).
   If the precise function being called is known, FUNC is its FUNCTION_DECL;
   otherwise, FUNC is 0.  */

/* On the Mac the return value is in D0 except for pascal routines, but
   pascal fakes this, since space has to be reserved ahead of time, which
   needs mods in the code proper. */
/* 68881 functions return floats in FP0. */

#define FUNCTION_VALUE(VALTYPE, FUNC)  \
	function_value( VALTYPE, FUNC )

#if 0
  gen_rtx (REG, TYPE_MODE (VALTYPE),  \
	   (TARGET_68881 &&  \
	    (GET_MODE_CLASS (TYPE_MODE (VALTYPE)) == MODE_FLOAT) ?  \
	    16 : 0))
#endif

/* The actual offset gets filled after assign_parms is run... */

#define FUNCTION_OUTGOING_VALUE(VALTYPE, FUNC) \
	function_outgoing_value( VALTYPE, FUNC )

#if 0
  (((FUNC) && current_function_is_pascal) ?  \
   gen_rtx (MEM, TYPE_MODE (VALTYPE),  \
            gen_rtx (PLUS, Pmode,  \
		     frame_pointer_rtx,  \
		     gen_rtx (CONST_INT, VOIDmode, 12345))) :  \
   FUNCTION_VALUE (VALTYPE, FUNC))
#endif

/* Define how to find the value returned by a library function
   assuming the value has mode MODE.  */

/* If we ever have any lib routines, they'll probably return in D0. */

#define LIBCALL_VALUE(MODE)  gen_rtx (REG, MODE, 0)

/* 1 if N is a possible register number for a function value.
/* This is d0 or fp0. */

#define FUNCTION_VALUE_REGNO_P(N)  \
  ((N) == 0 || (TARGET_68881 && (N) == 16))

/* Define this if PCC uses the nonreentrant convention for returning
   structure and union values.  */
/* MPW doesn't know diddly-squat about PCC. */

#undef PCC_STATIC_STRUCT_RETURN

/* 1 if N is a possible register number for function argument passing.
   On the 68000, no registers are used in this way.  */

#define FUNCTION_ARG_REGNO_P(N) 0

/* Define a data type for recording info about an argument list
   during the scan of that argument list.  This data type should
   hold all necessary information about the function itself
   and about the args processed so far, enough to enable macros
   such as FUNCTION_ARG to determine where the next arg should go.

   On the m68k, this is a single integer, which is a number of bytes
   of arguments scanned so far.  */

typedef struct {
	int size;
	int count;
	parameter *p;
} paramlist;

#define CUMULATIVE_ARGS paramlist

/* Initialize a variable CUM of type CUMULATIVE_ARGS
   for a call to a function whose data type is FNTYPE.
   For a library call, FNTYPE is 0.

   On the m68k, the offset starts at 0.  */

#define INIT_CUMULATIVE_ARGS(CUM,FNTYPE,PARA)	\
 ((CUM).size = 0, (CUM).count = 0, (CUM).p=(PARA) )

/* Update the data in CUM to advance over an argument
   of mode MODE and data type TYPE.
   (TYPE is null for libcalls where that information may not be available.)  */

#define FUNCTION_ARG_ADVANCE(CUM, MODE, TYPE, NAMED)	\
 ((CUM).size += ((MODE) != BLKmode			\
	    ? (GET_MODE_SIZE (MODE) + 3) & ~3	\
	    : (int_size_in_bytes (TYPE) + 3) & ~3),\
  (CUM).count +=1 )

/* Define where to put the arguments to a function.
   Value is zero to push the argument on the stack,
   or a hard register in which to store the argument.

   MODE is the argument's machine mode.
   TYPE is the data type of the argument (as a tree).
    This is null for libcalls where that information may
    not be available.
   CUM is a variable of type CUMULATIVE_ARGS which gives info about
    the preceding args and about the function being called.
   NAMED is nonzero if this argument is a named parameter
    (otherwise it is an extra parameter matching an ellipsis).  */

/* On the 68000 all args are pushed. */

#define FUNCTION_ARG(CUM, MODE, TYPE, NAMED) function_arg( &CUM, MODE, NAMED )

/* For an arg passed partly in registers and partly in memory,
   this is the number of registers used.
   For args passed entirely in registers or entirely in memory, zero.  */

#define FUNCTION_ARG_PARTIAL_NREGS(CUM, MODE, TYPE, NAMED) 0

/* This macro generates the assembly code for function entry.
   FILE is a stdio stream to output the code to.
   SIZE is an int: how many units of temporary storage to allocate.
   Refer to the array `regs_ever_live' to determine which registers
   to save; `regs_ever_live[I]' is nonzero if register number I
   is ever used in the function.  This macro is responsible for
   knowing which registers should not be saved even if used.  */

/* The ugliness here is thanks to the MPW assembler, which insists
   on dn/an register masks. */

#define FUNCTION_PROLOGUE(FILE, SIZE)  function_prologue (FILE, SIZE)

/* Output assembler code to FILE to increment profiler label # LABELNO
   for profiling a function entry.  */

#define FUNCTION_PROFILER(FILE, LABELNO)  \
  fprintf (FILE, "\tlea LP%d,a0\n\tjsr fn##count\n", LABELNO)

/* Output assembler code to FILE to initialize this source file's
   basic block profiling info, if that has not already been done.  */

#define FUNCTION_BLOCK_PROFILER(FILE, LABELNO)  \
  fprintf (FILE, "\ttst.l LPBX0\n\tbne LPI%d\n", LABELNO);  \
  fprintf (FILE, "\tpea LPBX0\n\tjsr bb##init\n\taddq.l #4,sp\n");  \
  fprintf (FILE, "LPI%d:\n", LABELNO);

/* Output assembler code to FILE to increment the entry-count for
   the BLOCKNO'th basic block in this source file.  */

#define BLOCK_PROFILER(FILE, BLOCKNO)	\
  fprintf (FILE, "\taddq.l #1,LPBX2+%d\n", 4 * BLOCKNO)

/* EXIT_IGNORE_STACK should be nonzero if, when returning from a function,
   the stack pointer does not matter.  The value is tested only in
   functions that have frame pointers.
   No definition is equivalent to always zero.  */

#define EXIT_IGNORE_STACK 1

/* This macro generates the assembly code for function exit,
   on machines that need it.  If FUNCTION_EPILOGUE is not defined
   then individual return instructions are generated for each
   return statement.  Args are same as for FUNCTION_PROLOGUE.

   The function epilogue should not depend on the current stack pointer!
   It should use the frame pointer only.  This is mandatory because
   of alloca; we also take advantage of it to omit stack adjustments
   before returning.  */

#define FUNCTION_EPILOGUE(FILE, SIZE) function_epilogue (FILE, SIZE)

/* If the memory address ADDR is relative to the frame pointer,
   correct it to be relative to the stack pointer instead.
   This is for when we don't use a frame pointer.
   ADDR should be a variable name.  */

#define FIX_FRAME_POINTER_ADDRESS(ADDR,DEPTH)  \
{ int offset = -1;							\
  rtx regs = stack_pointer_rtx;						\
  if (ADDR == frame_pointer_rtx)					\
    offset = 0;								\
  else if (GET_CODE (ADDR) == PLUS && XEXP (ADDR, 0) == frame_pointer_rtx \
	   && GET_CODE (XEXP (ADDR, 1)) == CONST_INT)			\
    offset = INTVAL (XEXP (ADDR, 1));					\
  else if (GET_CODE (ADDR) == PLUS && XEXP (ADDR, 0) == frame_pointer_rtx) \
    { rtx other_reg = XEXP (ADDR, 1);					\
      offset = 0;							\
      regs = gen_rtx (PLUS, Pmode, stack_pointer_rtx, other_reg); }	\
  else if (GET_CODE (ADDR) == PLUS && XEXP (ADDR, 1) == frame_pointer_rtx) \
    { rtx other_reg = XEXP (ADDR, 0);					\
      offset = 0;							\
      regs = gen_rtx (PLUS, Pmode, stack_pointer_rtx, other_reg); }	\
  else if (GET_CODE (ADDR) == PLUS					\
	   && GET_CODE (XEXP (ADDR, 0)) == PLUS				\
	   && XEXP (XEXP (ADDR, 0), 0) == frame_pointer_rtx		\
	   && GET_CODE (XEXP (ADDR, 1)) == CONST_INT)			\
    { rtx other_reg = XEXP (XEXP (ADDR, 0), 1);				\
      offset = INTVAL (XEXP (ADDR, 1));					\
      regs = gen_rtx (PLUS, Pmode, stack_pointer_rtx, other_reg); }	\
  else if (GET_CODE (ADDR) == PLUS					\
	   && GET_CODE (XEXP (ADDR, 0)) == PLUS				\
	   && XEXP (XEXP (ADDR, 0), 1) == frame_pointer_rtx		\
	   && GET_CODE (XEXP (ADDR, 1)) == CONST_INT)			\
    { rtx other_reg = XEXP (XEXP (ADDR, 0), 0);				\
      offset = INTVAL (XEXP (ADDR, 1));					\
      regs = gen_rtx (PLUS, Pmode, stack_pointer_rtx, other_reg); }	\
  if (offset >= 0)							\
    { int regno;							\
      extern char call_used_regs[];					\
      for (regno = 16; regno < FIRST_PSEUDO_REGISTER; regno++)		\
        if (regs_ever_live[regno] && ! call_used_regs[regno])		\
          offset += 12;							\
      for (regno = 0; regno < 16; regno++)				\
	if (regs_ever_live[regno] && ! call_used_regs[regno])		\
	  offset += 4;							\
      offset -= 4;							\
      ADDR = plus_constant (regs, offset + (DEPTH)); } }		\

/* Disable function cse even though it's good, because this screws up
   direct functions and I'm too lazy to put in the exact test everywhere. */

#define NO_FUNCTION_CSE


/* Addressing modes, and classification of registers for them.  */

#define HAVE_POST_INCREMENT

#define HAVE_PRE_DECREMENT

/* Macros to check register numbers against specific register classes.  */

/* These assume that REGNO is a hard or pseudo reg number.
   They give nonzero only if REGNO is a hard reg of the suitable class
   or a pseudo reg currently allocated to a suitable hard reg.
   Since they use reg_renumber, they are safe only once reg_renumber
   has been allocated, which happens in local-alloc.c.  */

#define REGNO_OK_FOR_INDEX_P(REGNO) \
((REGNO) < 16 || (unsigned) reg_renumber[REGNO] < 16)
#define REGNO_OK_FOR_BASE_P(REGNO) \
(((REGNO) ^ 010) < 8 || (unsigned) (reg_renumber[REGNO] ^ 010) < 8)
#define REGNO_OK_FOR_DATA_P(REGNO) \
((REGNO) < 8 || (unsigned) reg_renumber[REGNO] < 8)
#define REGNO_OK_FOR_FP_P(REGNO) \
(((REGNO) ^ 020) < 8 || (unsigned) (reg_renumber[REGNO] ^ 020) < 8)

/* Now macros that check whether X is a register and also,
   strictly, whether it is in a specified class.

   These macros are specific to the 68000, and may be used only
   in code for printing assembler insns and in conditions for
   define_optimization.  */

/* 1 if X is a data register.  */

#define DATA_REG_P(X) (REG_P (X) && REGNO_OK_FOR_DATA_P (REGNO (X)))

/* 1 if X is an fp register.  */

#define FP_REG_P(X) (REG_P (X) && REGNO_OK_FOR_FP_P (REGNO (X)))

/* 1 if X is an address register  */

#define ADDRESS_REG_P(X) (REG_P (X) && REGNO_OK_FOR_BASE_P (REGNO (X)))


/* Maximum number of registers that can appear in a valid memory address.  */

#define MAX_REGS_PER_ADDRESS 2

/* Recognize any constant value that is a valid address.  */

#define CONSTANT_ADDRESS_P(X)  \
  (GET_CODE (X) == LABEL_REF  /* || GET_CODE (X) == SYMBOL_REF  */	\
   || GET_CODE (X) == CONST_INT						\
   || GET_CODE (X) == CONST)

/*  CONSTANT_P (X)  */

/* Nonzero if the constant value X is a legitimate general operand.
   It is given that X satisfies CONSTANT_P or is a CONST_DOUBLE.  */

#define LEGITIMATE_CONSTANT_P(X) 1

/* The macros REG_OK_FOR..._P assume that the arg is a REG rtx
   and check its validity for a certain class.
   We have two alternate definitions for each of them.
   The usual definition accepts all pseudo regs; the other rejects
   them unless they have been allocated suitable hard regs.
   The symbol REG_OK_STRICT causes the latter definition to be used.

   Most source files want to accept pseudo regs in the hope that
   they will get allocated to the class that the insn wants them to be in.
   Source files for reload pass need to be strict.
   After reload, it makes no difference, since pseudo regs have
   been eliminated by then.  */

#ifndef REG_OK_STRICT

/* Nonzero if X is a hard reg that can be used as an index
   or if it is a pseudo reg.  */
#define REG_OK_FOR_INDEX_P(X) ((REGNO (X) ^ 020) >= 8)
/* Nonzero if X is a hard reg that can be used as a base reg
   or if it is a pseudo reg.  */
#define REG_OK_FOR_BASE_P(X) ((REGNO (X) & ~027) != 0)

#else

/* Nonzero if X is a hard reg that can be used as an index.  */
#define REG_OK_FOR_INDEX_P(X) REGNO_OK_FOR_INDEX_P (REGNO (X))
/* Nonzero if X is a hard reg that can be used as a base reg.  */
#define REG_OK_FOR_BASE_P(X) REGNO_OK_FOR_BASE_P (REGNO (X))

#endif

/* GO_IF_LEGITIMATE_ADDRESS recognizes an RTL expression
   that is a valid memory address for an instruction.
   The MODE argument is the machine mode for the MEM expression
   that wants to use this address.

   The other macros defined here are used only in GO_IF_LEGITIMATE_ADDRESS.  */

#define GO_IF_INDEXING(X, ADDR)	\
{ if (GET_CODE (X) == PLUS && LEGITIMATE_INDEX_P (XEXP (X, 0)))		\
   { if (GET_CODE (XEXP (X, 1)) == LABEL_REF) goto ADDR;		\
     if (GET_CODE (XEXP (X, 1)) == REG && REG_OK_FOR_BASE_P (XEXP (X, 1)))  \
       goto ADDR; }  \
  if (GET_CODE (X) == PLUS && LEGITIMATE_INDEX_P (XEXP (X, 1)))		\
   { if (GET_CODE (XEXP (X, 0)) == LABEL_REF) goto ADDR;		\
     if (GET_CODE (XEXP (X, 0)) == REG && REG_OK_FOR_BASE_P (XEXP (X, 0)))  \
       goto ADDR; } }

#ifdef MPW_C

/* MPW C runs out of space compiling large macros, so use simplified forms. */

#define LEGITIMATE_INDEX_REG_P(X)   \
  ((GET_CODE (X) == REG && REG_OK_FOR_INDEX_P (X)))

#define LEGITIMATE_INDEX_P(X)   \
   (LEGITIMATE_INDEX_REG_P (X))

#else

#define LEGITIMATE_INDEX_REG_P(X)   \
  ((GET_CODE (X) == REG && REG_OK_FOR_INDEX_P (X))      \
   || (GET_CODE (X) == SIGN_EXTEND			\
       && GET_CODE (XEXP (X, 0)) == REG			\
       && GET_MODE (XEXP (X, 0)) == HImode		\
       && REG_OK_FOR_INDEX_P (XEXP (X, 0))))

#define LEGITIMATE_INDEX_P(X)   \
   (LEGITIMATE_INDEX_REG_P (X)  \
    || (TARGET_68020 && GET_CODE (X) == MULT		\
	&& LEGITIMATE_INDEX_REG_P (XEXP (X, 0))		\
	&& GET_CODE (XEXP (X, 1)) == CONST_INT		\
	&& (INTVAL (XEXP (X, 1)) == 2			\
	    || INTVAL (XEXP (X, 1)) == 4		\
	    || INTVAL (XEXP (X, 1)) == 8)))

#endif /* MPW_C */

#define GO_IF_LEGITIMATE_ADDRESS(MODE, X, ADDR)  \
{ if (CONSTANT_ADDRESS_P (X)						\
      || (GET_CODE (X) == SYMBOL_REF)  \
      || (GET_CODE (X) == REG && REG_OK_FOR_BASE_P (X))			\
      || ((GET_CODE (X) == PRE_DEC || GET_CODE (X) == POST_INC)		\
          && REG_P (XEXP (X, 0))					\
          && REG_OK_FOR_BASE_P (XEXP (X, 0)))				\
      || (GET_CODE (X) == PLUS						\
          && REG_P (XEXP (X, 0)) && REG_OK_FOR_BASE_P (XEXP (X, 0))	\
          && GET_CODE (XEXP (X, 1)) == CONST_INT			\
          && ((unsigned) INTVAL (XEXP (X, 1)) + 0x8000) < 0x10000)) goto ADDR;\
  GO_IF_INDEXING (X, ADDR);						\
  if (GET_CODE (X) == PLUS)						\
    { if (GET_CODE (XEXP (X, 1)) == CONST_INT				\
          && (unsigned) INTVAL (XEXP (X, 1)) + 0x80 < 0x100)		\
	   { rtx go_temp = XEXP (X, 0); GO_IF_INDEXING (go_temp, ADDR); }  \
      if (GET_CODE (XEXP (X, 0)) == CONST_INT				\
	  && (unsigned) INTVAL (XEXP (X, 0)) + 0x80 < 0x100)		\
	   { rtx go_temp = XEXP (X, 1); GO_IF_INDEXING (go_temp, ADDR); } } }

/* Try machine-dependent ways of modifying an illegitimate address
   to be legitimate.  If we find one, return the new, valid address.
   This macro is used in only one place: `memory_address' in explow.c.

   OLDX is the address as it was before break_out_memory_refs was called.
   In some cases it is useful to look at this to decide what needs to be done.

   MODE and WIN are passed so that this macro can use
   GO_IF_LEGITIMATE_ADDRESS.

   It is always safe for this macro to do nothing.  It exists to recognize
   opportunities to optimize the output.

   For the 68000, we handle X+REG by loading X into a register R and
   using R+REG.  R will go in an address reg and indexing will be used.
   However, if REG is a broken-out memory address or multiplication,
   nothing needs to be done because REG can certainly go in an address reg.  */

#define LEGITIMIZE_ADDRESS(X,OLDX,MODE,WIN)   \
{ register int ch = (X) != (OLDX);					\
  if (GET_CODE (X) == PLUS)						\
    { if (GET_CODE (XEXP (X, 0)) == MULT)				\
	ch = 1, XEXP (X, 0) = force_operand (XEXP (X, 0), 0);		\
      if (GET_CODE (XEXP (X, 1)) == MULT)				\
	ch = 1, XEXP (X, 1) = force_operand (XEXP (X, 1), 0);		\
      if (ch && GET_CODE (XEXP (X, 1)) == REG				\
	  && GET_CODE (XEXP (X, 0)) == REG)				\
	return X;							\
      if (ch) { GO_IF_LEGITIMATE_ADDRESS (MODE, X, WIN); }		\
      if (GET_CODE (XEXP (X, 0)) == REG					\
	       || (GET_CODE (XEXP (X, 0)) == SIGN_EXTEND		\
		   && GET_CODE (XEXP (XEXP (X, 0), 0)) == REG		\
		   && GET_MODE (XEXP (XEXP (X, 0), 0)) == HImode))	\
	{ register rtx temp = gen_reg_rtx (Pmode);			\
	  register rtx val = force_operand (XEXP (X, 1), 0);		\
	  emit_move_insn (temp, val);					\
	  XEXP (X, 1) = temp;						\
	  return X; }							\
      else if (GET_CODE (XEXP (X, 1)) == REG				\
	       || (GET_CODE (XEXP (X, 1)) == SIGN_EXTEND		\
		   && GET_CODE (XEXP (XEXP (X, 1), 0)) == REG		\
		   && GET_MODE (XEXP (XEXP (X, 1), 0)) == HImode))	\
	{ register rtx temp = gen_reg_rtx (Pmode);			\
	  register rtx val = force_operand (XEXP (X, 0), 0);		\
	  emit_move_insn (temp, val);					\
	  XEXP (X, 0) = temp;						\
	  return X; }}}

/* Go to LABEL if ADDR (a legitimate address expression)
   has an effect that depends on the machine mode it is used for.
   On the 68000, only predecrement and postincrement address depend thus
   (the amount of decrement or increment being the length of the operand).  */

#define GO_IF_MODE_DEPENDENT_ADDRESS(ADDR,LABEL)	\
 if (GET_CODE (ADDR) == POST_INC || GET_CODE (ADDR) == PRE_DEC) goto LABEL

/* Specify the machine mode that this machine uses
   for the index in the tablejump instruction.  */
#define CASE_VECTOR_MODE HImode

/* Define this if the tablejump instruction expects the table
   to contain offsets from the address of the table.
   Do not define this if the table should contain absolute addresses.  */
#undef CASE_VECTOR_PC_RELATIVE

/* Specify the tree operation to be used to convert reals to integers.  */
#define IMPLICIT_FIX_EXPR FIX_ROUND_EXPR

/* This is the kind of divide that is easiest to do in the general case.  */
#define EASY_DIV_EXPR TRUNC_DIV_EXPR

/* Define this as 1 if `char' should by default be signed; else as 0.  */
#define DEFAULT_SIGNED_CHAR 1

/* Max number of bytes we can move from memory to memory
   in one reasonably fast instruction.  */
#define MOVE_MAX 4

/* Define this if zero-extension is slow (more than one real instruction).  */
#define SLOW_ZERO_EXTEND

/* Nonzero if access to memory by bytes is slow and undesirable.  */
#define SLOW_BYTE_ACCESS 0

/* Define if shifts truncate the shift count
   which implies one can omit a sign-extension or zero-extension
   of a shift count.  */
#define SHIFT_COUNT_TRUNCATED

/* Value is 1 if truncating an integer of INPREC bits to OUTPREC bits
   is done just by pretending it is already truncated.  */
#define TRULY_NOOP_TRUNCATION(OUTPREC, INPREC) 1

/* We assume that the store-condition-codes instructions store 0 for false
   and some other value for true.  This is the value stored for true.  */

#define STORE_FLAG_VALUE -1

/* When a prototype says `char' or `short', really pass an `int'.  */
/* This has to be selectively modified in the code, so as to accommodate
   Pascal functions that sometimes need only shorts to be passed. */

#define PROMOTE_PROTOTYPES

/* Specify the machine mode that pointers have.
   After generation of rtl, the compiler makes no further distinction
   between pointers and any other objects of this machine mode.  */
#define Pmode SImode

/* A function address in a call instruction
   is a byte address (for indexing purposes)
   so give the MEM rtx a byte's mode.  */
#define FUNCTION_MODE QImode

/* Compute the cost of computing a constant rtl expression RTX
   whose rtx-code is CODE.  The body of this macro is a portion
   of a switch statement.  If the code is computed here,
   return it with a return statement.  Otherwise, break from the switch.  */

#define CONST_COSTS(RTX,CODE) \
  case CONST_INT:						\
    /* Constant zero is super cheap due to clr instruction.  */	\
    if (RTX == const0_rtx) return 0;				\
    if ((unsigned) INTVAL (RTX) < 077) return 1;		\
  case CONST:							\
  case LABEL_REF:						\
  case SYMBOL_REF:						\
    return 3;							\
  case CONST_DOUBLE:						\
    return 5;

/* Tell final.c how to eliminate redundant test instructions.  */

/* Here we define machine-dependent flags and fields in cc_status
   (see `conditions.h').  */

/* Set if the cc value is actually in the 68881, so a floating point
   conditional branch must be output.  */
#define CC_IN_68881 04000

/* Store in cc_status the expressions
   that the condition codes will describe
   after execution of an instruction whose pattern is EXP.
   Do not alter them if the instruction would not alter the cc's.  */

/* On the 68000, all the insns to store in an address register
   fail to set the cc's.  However, in some cases these instructions
   can make it possibly invalid to use the saved cc's.  In those
   cases we clear out some or all of the saved cc's so they won't be used.  */

#define NOTICE_UPDATE_CC(EXP, INSN) \
{								\
  if (GET_CODE (EXP) == SET)				\
    { if (ADDRESS_REG_P (SET_DEST (EXP)))			\
	{ if (cc_status.value1					\
	      && reg_overlap_mentioned_p (SET_DEST (EXP), cc_status.value1)) \
	    cc_status.value1 = 0;				\
	  if (cc_status.value2					\
	      && reg_overlap_mentioned_p (SET_DEST (EXP), cc_status.value2)) \
	    cc_status.value2 = 0; }				\
      else if (!FP_REG_P (SET_DEST (EXP))			\
	       && SET_DEST (EXP) != cc0_rtx			\
	       && (FP_REG_P (SET_SRC (EXP))			\
		   || GET_CODE (SET_SRC (EXP)) == FIX		\
		   || GET_CODE (SET_SRC (EXP)) == FLOAT_TRUNCATE \
		   || GET_CODE (SET_SRC (EXP)) == FLOAT_EXTEND)) \
	{ CC_STATUS_INIT; }					\
      /* A pair of move insns doesn't produce a useful overall cc.  */ \
      else if (!FP_REG_P (SET_DEST (EXP))			\
	       && !FP_REG_P (SET_SRC (EXP))			\
	       && GET_MODE_SIZE (GET_MODE (SET_SRC (EXP))) > 4	\
	       && (GET_CODE (SET_SRC (EXP)) == REG		\
		   || GET_CODE (SET_SRC (EXP)) == MEM		\
		   || GET_CODE (SET_SRC (EXP)) == CONST_DOUBLE)) \
	{ CC_STATUS_INIT; }					\
      else if (GET_CODE (SET_SRC (EXP)) == CALL)		\
	{ CC_STATUS_INIT; }					\
      else if (XEXP (EXP, 0) != pc_rtx)				\
	{ cc_status.flags = 0;					\
	  cc_status.value1 = XEXP (EXP, 0);			\
	  cc_status.value2 = XEXP (EXP, 1); } }			\
  else if (GET_CODE (EXP) == PARALLEL				\
	   && GET_CODE (XVECEXP (EXP, 0, 0)) == SET)		\
    {								\
      if (ADDRESS_REG_P (XEXP (XVECEXP (EXP, 0, 0), 0)))	\
	CC_STATUS_INIT;						\
      else if (XEXP (XVECEXP (EXP, 0, 0), 0) != pc_rtx)		\
	{ cc_status.flags = 0;					\
	  cc_status.value1 = XEXP (XVECEXP (EXP, 0, 0), 0);	\
	  cc_status.value2 = XEXP (XVECEXP (EXP, 0, 0), 1); } }	\
  else CC_STATUS_INIT;						\
  if (cc_status.value2 != 0					\
      && ADDRESS_REG_P (cc_status.value2)			\
      && GET_MODE (cc_status.value2) == QImode)			\
    CC_STATUS_INIT;						\
  if (cc_status.value1 && GET_CODE (cc_status.value1) == REG	\
      && cc_status.value2					\
      && reg_overlap_mentioned_p (cc_status.value1, cc_status.value2))	\
    cc_status.value2 = 0;					\
  if (((cc_status.value1 && FP_REG_P (cc_status.value1))	\
       || (cc_status.value2 && FP_REG_P (cc_status.value2)))	\
      && !((cc_status.value1 && 0)	\
	   || (cc_status.value2 && 0)))  \
    cc_status.flags = CC_IN_68881; }

#define OUTPUT_JUMP(NORMAL, FLOAT, NO_OV)  \
{ if (cc_prev_status.flags & CC_IN_68881)			\
    return FLOAT;						\
  if (cc_prev_status.flags & CC_NO_OVERFLOW)			\
    return NO_OV;						\
  return NORMAL; }

/* Control the assembler format that we output.  */

#define ASM_IDENTIFY_GCC(FILE)  \
{ extern char *version_string;  \
  fprintf((FILE), "%c Compiled by MPW GCC %s\n", TARGET_GAS?'#':';', version_string);  }

/* Dump out all sorts of random header stuff that MPW wants. */

#define ASM_FILE_START(FILE)	\
{  \
  ASM_SEGMENT_NAME(FILE, segment_name);  \
  if (TARGET_68020 && !TARGET_GAS)  \
    fprintf (FILE, "\tMACHINE MC68020\n");  \
  if (!TARGET_68020 && !TARGET_GAS)  \
    fprintf (FILE, "\tMACHINE MC68000\n");  \
  if (TARGET_68881 && !TARGET_GAS)  \
    fprintf (FILE, "\tMC68881\n");  \
  if (TARGET_32BITDATA && TARGET_68020 && !TARGET_GAS)  \
    fprintf (FILE, "\tFORWARD LONG\n");  \
  if (!TARGET_GAS) \
    { fprintf (FILE, "\tSTRING ASIS\n");  \
      fprintf (FILE, "\tCASE ON\n"); } \
}

/* Output to assembler file text saying following lines
   may contain character constants, extra white space, comments, etc.  */

#define ASM_APP_ON TARGET_GAS?"#APP\n":"; begin asm code\n"

/* Output to assembler file text saying following lines
   no longer contain unusual constructs.  */

#define ASM_APP_OFF TARGET_GAS?"#NOAPP\n":"; end asm code\n"

/* Output before read-only data.  */

#define TEXT_SECTION_ASM_OP TARGET_GAS?"\t.text":"\tCODE"

/* Output before writable data.  */

#define DATA_SECTION_ASM_OP TARGET_GAS?"\t.data":"\tDATA"

/* Force all data thingies and variables to be data, except for strings
   glued onto the end of function modules. */

#define SELECT_SECTION(DECL)  \
  if (TARGET_B && (TREE_CODE (DECL) == STRING_CST) && current_function_decl)  \
    ;  \
  else  \
    data_section ();

#define SELECT_RTX_SECTION(MODE,X) data_section ()

/* The MPW proc declaration autoswitches to text section, so we need this
   to tell GCC that this has happened. */

#define EXTRA_SECTION_FUNCTIONS  \
  void force_text_section () { in_section = in_text; }

/* How to refer to registers in assembler output.
   This sequence is indexed by compiler's hard-register-number (see above).  */

#define REGISTER_NAMES \
{"d0", "d1", "d2", "d3", "d4", "d5", "d6", "d7",	\
 "a0", "a1", "a2", "a3", "a4", "a5", "a6", "sp",	\
 "fp0", "fp1", "fp2", "fp3", "fp4", "fp5", "fp6", "fp7", }

/* How to renumber registers for dbx and gdb. Not used... */

#define DBX_REGISTER_NUMBER(REGNO) (REGNO)

/* Note that PROC in MPW asm automatically declares CODE, so we
   have to change the assumed section without writing the directive. */

#define ASM_PREDECLARE_FUNCTION_NAME(FILE,NAME,DECL)  \
{  \
  if (!TARGET_GAS) { \
        assemble_name (FILE, NAME);  \
        fprintf (FILE, "%% \tPROC\n");  \
        force_text_section ();  \
        } \
  import_undefined (FILE);  \
  text_section();  \
}

#define ASM_DECLARE_FUNCTION_NAME(FILE,NAME,DECL)  \
{  if (TARGET_GAS) {\
  if ((DECL) && TREE_PUBLIC (DECL)) {  \
    fprintf (FILE, "\t.globl ");  \
  assemble_name (FILE, NAME);  \
  fprintf (FILE, "\n"); } \
} else { \
  if (TARGET_FX30)  \
    fprintf (FILE, "\tIMPORT _StaticDataArea\n");  \
  if ((DECL) && TREE_PUBLIC (DECL))   \
    fprintf (FILE, "\tEXPORT ");  \
  else  \
    fprintf (FILE, "\tENTRY ");  \
  assemble_name (FILE, NAME);  \
  fprintf (FILE, "\n");  }\
  ASM_OUTPUT_LABEL (FILE, NAME);  \
}
/*
  fprintf (FILE, "%s \tPROC\n", (strcmp(NAME, "1main") == 0 ? "%" : "");  \
*/

#define ASM_DECLARE_VARIABLE_NAME(FILE,NAME,PUBLIC)  \
{  if (TARGET_GAS) {\
  if (PUBLIC) {  \
    fprintf (FILE, "\t.globl ");  \
  assemble_name (FILE, NAME);  \
  fprintf (FILE, "\n"); }  \
} else {\
  assemble_name (FILE, NAME);  \
  fprintf (FILE, "%% \tPROC\n");  \
  force_text_section ();  \
  import_undefined (FILE);  \
  data_section ();  \
  if (PUBLIC)   \
    fprintf (FILE, "\tEXPORT ");  \
  else  \
    fprintf (FILE, "\tENTRY ");  \
  assemble_name (FILE, NAME);  \
  fprintf (FILE, "\n");  }\
}

#define ASM_DECLARE_CONSTANT_NAME(FILE,NUM,EXP)  \
{  char label[256];\
  ASM_GENERATE_INTERNAL_LABEL (label, "LC", NUM);  \
  if (TARGET_B && TREE_CODE(EXP) == STRING_CST && current_function_decl)  \
    {  \
    }  \
  else if (!TARGET_GAS)  \
    {  \
      assemble_name (FILE, label);  \
      fprintf (FILE, "%% \tPROC\n");  \
      force_text_section ();  \
    }  \
}

#define ASM_ENTRY_CONSTANT_NAME(FILE,NUM,EXP)  \
{  char label[256];\
  ASM_GENERATE_INTERNAL_LABEL (label, "LC", NUM);  \
  if (TARGET_B && TREE_CODE(EXP) == STRING_CST && current_function_decl)  \
    {  \
    }  \
  else if (!TARGET_GAS)  \
    {  \
      fprintf (FILE, "\tENTRY ");  \
      assemble_name (FILE, label);  \
      fprintf (FILE, "\n");  \
    }  \
}

#define OUTPUT_LABELED_CONSTANT(asm_out_file,const_labelno,exp)  \
  if (! (TARGET_B  && TREE_CODE (exp) == STRING_CST && current_function_decl)) {  \
    ASM_OUTPUT_INTERNAL_LABEL (asm_out_file, "LC", const_labelno);  \
    output_constant (exp,  \
		   (TREE_CODE (exp) == STRING_CST  \
		    ? TREE_STRING_LENGTH (exp)  \
		    : int_size_in_bytes (TREE_TYPE (exp))));  \
  }

#define REMEMBER_CONSTANT_LABEL(lbl,exp)  \
  if (TARGET_B  && TREE_CODE (exp) == STRING_CST && current_function_decl) \
    {  \
      ASM_FUNCTIONIFY_NAME (lbl);  \
      record_a_string (lbl, TREE_STRING_POINTER (exp));  \
      }

/* Exportation is done by general declaration macros. */

#define ASM_GLOBALIZE_LABEL(FILE,NAME)

#define ASM_OUTPUT_EXTERNAL(FILE,DECL,NAME)  \
  import_a_declared_name(DECL);

/* This is how to output the definition of a user-level label named NAME,
   such as the label on a static function or variable NAME.  */

#define ASM_OUTPUT_LABEL(FILE,NAME)	\
{ assemble_name (FILE, NAME); fputs (":\n", FILE); }

/* This is how to output a reference to a user-level label named NAME.
   `assemble_name' uses this.  */

/* User-level labels (almost) always have one extra character that should
   be ignored.  They may also conflict with MPW Asm register names. */

#define ASM_OUTPUT_LABELREF(FILE,NAME)  \
{  \
  extern char *avoid_mpw_register_name();  \
  char *tname = avoid_mpw_register_name (NAME);  \
  if ('0' <= tname[0] && tname[0] <= '9')  \
   fprintf (FILE, (TARGET_GAS&&(tname[1]!=' '))?"_%s":"%s", tname+1);  \
  else  \
   fprintf (FILE, (TARGET_GAS&&(tname[0]!=' '))?"_%s":"%s", tname);  \
}

/* This is how to output an internal numbered label where
   PREFIX is the class of label and NUM is the number within the class.  */

#define ASM_OUTPUT_INTERNAL_LABEL(FILE,PREFIX,NUM)	\
  fprintf (FILE,  TARGET_GAS ? "_%s#%d:\n":"%s#%d:\n", PREFIX, NUM)

/* This is how to store into the string LABEL
   the symbol_ref name of an internal numbered label where
   PREFIX is the class of label and NUM is the number within the class.
   This is suitable for output with `assemble_name'.  */

#define ASM_GENERATE_INTERNAL_LABEL(LABEL,PREFIX,NUM)	\
  sprintf (LABEL, "2%s#%d", PREFIX, NUM)

/* This is how to output an assembler line defining a `float' constant.  */

#define ASM_OUTPUT_FLOAT(FILE,VALUE)  \
  fprintf (FILE, TARGET_GAS?"\t.float %.9g\n":"\tDC.S \"%.9g\"\n", (VALUE))

#define ASM_OUTPUT_DOUBLE(FILE,VALUE)  \
  fprintf (FILE, TARGET_GAS?"\t.double %.20g\n":"\tDC.D \"%.20g\"\n", (VALUE))

#define ASM_OUTPUT_LONG_DOUBLE(FILE,VALUE)  \
  output_mpw_long_double (FILE, VALUE)

/* This is how to output an assembler line defining an `int' constant.  */

#define ASM_OUTPUT_INT(FILE,VALUE)  \
( fprintf (FILE, TARGET_GAS?"\t.long ":"\tDC.L "),			\
  output_addr_const (FILE, (VALUE)),		\
  fprintf (FILE, "\n"))

/* Likewise for `char' and `short' constants.  */

#define ASM_OUTPUT_SHORT(FILE,VALUE)  \
( fprintf (FILE, TARGET_GAS?"\t.word ":"\tDC.W "),			\
  output_addr_const (FILE, (VALUE)),		\
  fprintf (FILE, "\n"))

#define ASM_OUTPUT_CHAR(FILE,VALUE)  \
( fprintf (FILE, TARGET_GAS?"\t.byte ":"\tDC.B "),			\
  output_addr_const (FILE, (VALUE)),		\
  fprintf (FILE, "\n"))

/* This is how to output an assembler line for a numeric constant byte.  */

#define ASM_OUTPUT_BYTE(FILE,VALUE)  \
  fprintf (FILE, TARGET_GAS?"\t.byte $%x\n":"\tDC.B $%x\n", (VALUE))

/* This is how to output an insn to push a register on the stack.
   It need not be very fast code.  */

#define ASM_OUTPUT_REG_PUSH(FILE,REGNO)  \
  fprintf (FILE, "\tmove.l %s,-(sp)\n", reg_names[REGNO])

/* This is how to output an insn to pop a register from the stack.
   It need not be very fast code.  */

#define ASM_OUTPUT_REG_POP(FILE,REGNO)  \
  fprintf (FILE, "\tmove.l (sp)+,%s\n", reg_names[REGNO])

/* This is how to output an element of a case-vector that is absolute.
   (The 68000 does not use such vectors,
   but we must define this macro anyway.)  */

#define ASM_OUTPUT_ADDR_VEC_ELT(FILE, VALUE)  \
  fprintf (FILE, TARGET_GAS?"\t.long L#%d\n":"\tDC.L L#%d\n", VALUE)

/* This is how to output an element of a case-vector that is relative.  */

#define ASM_OUTPUT_ADDR_DIFF_ELT(FILE, VALUE, REL)  \
  fprintf (FILE, TARGET_GAS?"\t.word L#%d-L#%d\n":"\tDC.W L#%d-L#%d\n", VALUE, REL)

/* This is how to output an assembler line
   that says to advance the location counter
   to a multiple of 2**LOG bytes.  */

#define ASM_OUTPUT_ALIGN(FILE,LOG)      \
  if ((LOG) == 1)			\
    fprintf (FILE, TARGET_GAS?"\t.even\n":"\tALIGN 2\n");	\
  else if ((LOG) != 0)			\
    abort ();

/* A "skip" in MPW is better done by filling with zeros. */

#define ASM_OUTPUT_SKIP(FILE,SIZE)  \
  fprintf (FILE, TARGET_GAS?"\t.skip %d\n":"\tDCB.B %d,0\n", (SIZE))

/* This says how to output an assembler line
   to define a global common symbol.  */

#define ASM_OUTPUT_COMMON(FILE, NAME, SIZE, ROUNDED)  \
do {  if (TARGET_GAS) fputs ("\t.comm ", (FILE));			\
  assemble_name (FILE, NAME);  \
  if (TARGET_GAS) fprintf ((FILE), ",%u\n", (ROUNDED)); \
  else { fprintf (FILE, "%% \tPROC\n");  \
  force_text_section ();  \
  data_section ();  \
  fprintf (FILE, "\tEXPORT ");  \
  assemble_name (FILE, NAME);  \
  fprintf (FILE, "\n");  \
  assemble_name (FILE, NAME);  \
  fprintf (FILE, ":\tDS.B %d\n", ROUNDED);  \
  fprintf (FILE, "\tENDP\t; ");  \
  assemble_name (FILE, NAME);  \
  fprintf (FILE, "\n"); } \
} while (0)

/* This says how to output an assembler line
   to define a local common symbol.  */

#define ASM_OUTPUT_LOCAL(FILE, NAME, SIZE, ROUNDED)  \
do {  if (TARGET_GAS) fputs ("\t.lcomm ", (FILE));			\
  assemble_name (FILE, NAME);  \
  if (TARGET_GAS) fprintf ((FILE), ",%u\n", (ROUNDED)); \
  else {   fprintf (FILE, "%% \tPROC\n");  \
  force_text_section ();  \
  data_section ();  \
  fprintf (FILE, "\tENTRY ");  \
  assemble_name (FILE, NAME);  \
  fprintf (FILE, "\n");  \
  assemble_name (FILE, NAME);  \
  fprintf (FILE, ":\tDS.B %d\n", ROUNDED);  \
  fprintf (FILE, "\tENDP\t; ");  \
  assemble_name (FILE, NAME);  \
  fprintf (FILE, "\n"); } \
} while (0)

/* Store in OUTPUT a string (made with alloca) containing
   an assembler-name for a local static variable named NAME.
   LABELNO is an integer which is different for each call.  */

#define ASM_FORMAT_PRIVATE_NAME(OUTPUT, NAME, LABELNO)	\
( (OUTPUT) = (char *) alloca (strlen ((NAME)) + 12),	\
  sprintf ((OUTPUT), "%s%s%d", (NAME), TARGET_GAS?".":"___", (LABELNO)))

/* Modify symbol names to say what sort they are.  These can be destructive
   to the name string since they are supposed to be permanent. */

#define ASM_FUNCTIONIFY_NAME(NAME)  ((NAME)[0] = '1')

#define ASM_DATIFY_NAME(NAME)  ((NAME)[0] = '2')

/* Define the parentheses used to group arithmetic operations
   in assembler code.  */

#define ASM_OPEN_PAREN "("
#define ASM_CLOSE_PAREN ")"

/* Define results of standard character escape sequences.  */
#define TARGET_BELL 007
#define TARGET_BS 010
#define TARGET_TAB 011
#define TARGET_NEWLINE 13
#define TARGET_VT 013
#define TARGET_FF 014
#define TARGET_CR 10

/* Print operand X (an rtx) in assembler syntax to file FILE.
   CODE is a letter or dot (`z' in `%z0') or 0 if no letter was specified.
   For `%' followed by punctuation, CODE is the punctuation and X is null.

   On the 68000, we use several CODE characters:
   '#' for an immediate operand prefix.
   '.' for dot needed in Motorola-style opcode names.
   '-' for an operand pushing on the stack:
       sp@-, -(sp) or -(%sp) depending on the style of syntax.
   '+' for an operand pushing on the stack:
       sp@+, (sp)+ or (%sp)+ depending on the style of syntax.
   '@' for a reference to the top word on the stack:
       sp@, (sp) or (%sp) depending on the style of syntax.

   'b' for byte insn (no effect, on the Sun; this is for the ISI).
   'd' to force memory addressing to be absolute, not relative.
*/

#define PRINT_OPERAND_PUNCT_VALID_P(CODE)				\
  ((CODE) == '.' || (CODE) == '#' || (CODE) == '-'			\
   || (CODE) == '+' || (CODE) == '@')

/* Note absence of float operands - they are always referred to by address
   and never appear directly in instructions (is this good?). */

#define PRINT_OPERAND(FILE, X, CODE) \
 { if (CODE=='.') \
        { \
        if (!TARGET_GAS) putc('.',FILE); \
        } \
  else if (CODE == '-') fprintf (FILE, TARGET_GAS?"sp@-":"-(sp)"); \
  else if (CODE == '+') fprintf (FILE, TARGET_GAS?"sp@+":"(sp)+"); \
  else if (CODE == '@') fprintf (FILE, TARGET_GAS?"sp@":"(sp)");	\
  else if (GET_CODE (X) == REG)						\
    fprintf (FILE, "%s", reg_names[REGNO (X)]);				\
  else if (GET_CODE (X) == MEM)						\
    output_address (XEXP (X, 0));					\
  else if (CODE == 'b')  \
    output_addr_const (FILE, X);  \
  else if (GET_CODE (X) == CONST_DOUBLE)  \
    {  \
      putc ('#', FILE);  \
      if (CODE == 'f')  \
	output_mpw_float_as_int (FILE, X);  \
      else  \
	output_mpw_float (FILE, X);  \
    }  \
  else { putc ('#', FILE); output_addr_const (FILE, X); }}

#define PRINT_OPERAND_ADDRESS_GAS(FILE, ADDR)  \
  switch (GET_CODE (addr))						\
    {									\
    case REG:								\
      fprintf (FILE, "%s@", reg_names[REGNO (addr)]);			\
      break;								\
    case PRE_DEC:							\
      fprintf (FILE, "%s@-", reg_names[REGNO (XEXP (addr, 0))]);	\
      break;								\
    case POST_INC:							\
      fprintf (FILE, "%s@+", reg_names[REGNO (XEXP (addr, 0))]);	\
      break;								\
    case PLUS:								\
      reg1 = 0;	reg2 = 0;						\
      ireg = 0;	breg = 0;						\
      offset = 0;							\
      if (CONSTANT_ADDRESS_P (XEXP (addr, 0)))				\
	{								\
	  offset = XEXP (addr, 0);					\
	  addr = XEXP (addr, 1);					\
	}								\
      else if (CONSTANT_ADDRESS_P (XEXP (addr, 1)))			\
	{								\
	  offset = XEXP (addr, 1);					\
	  addr = XEXP (addr, 0);					\
	}								\
      if (GET_CODE (addr) != PLUS) ;					\
      else if (GET_CODE (XEXP (addr, 0)) == SIGN_EXTEND)		\
	{								\
	  reg1 = XEXP (addr, 0);					\
	  addr = XEXP (addr, 1);					\
	}								\
      else if (GET_CODE (XEXP (addr, 1)) == SIGN_EXTEND)		\
	{								\
	  reg1 = XEXP (addr, 1);					\
	  addr = XEXP (addr, 0);					\
	}								\
      else if (GET_CODE (XEXP (addr, 0)) == MULT)			\
	{								\
	  reg1 = XEXP (addr, 0);					\
	  addr = XEXP (addr, 1);					\
	}								\
      else if (GET_CODE (XEXP (addr, 1)) == MULT)			\
	{								\
	  reg1 = XEXP (addr, 1);					\
	  addr = XEXP (addr, 0);					\
	}								\
      else if (GET_CODE (XEXP (addr, 0)) == REG)			\
	{								\
	  reg1 = XEXP (addr, 0);					\
	  addr = XEXP (addr, 1);					\
	}								\
      else if (GET_CODE (XEXP (addr, 1)) == REG)			\
	{								\
	  reg1 = XEXP (addr, 1);					\
	  addr = XEXP (addr, 0);					\
	}								\
      if (GET_CODE (addr) == REG || GET_CODE (addr) == MULT		\
	  || GET_CODE (addr) == SIGN_EXTEND)				\
	{ if (reg1 == 0) reg1 = addr; else reg2 = addr; addr = 0; }	\
      if (offset != 0) { if (addr != 0) abort (); addr = offset; }	\
      if ((reg1 && (GET_CODE (reg1) == SIGN_EXTEND			\
		    || GET_CODE (reg1) == MULT))			\
	  || (reg2 != 0 && REGNO_OK_FOR_BASE_P (REGNO (reg2))))		\
	{ breg = reg2; ireg = reg1; }					\
      else if (reg1 != 0 && REGNO_OK_FOR_BASE_P (REGNO (reg1)))		\
	{ breg = reg1; ireg = reg2; }					\
      if (ireg != 0 && breg == 0 && GET_CODE (addr) == LABEL_REF)	\
        { int scale = 1;						\
	  if (GET_CODE (ireg) == MULT)					\
	    { scale = INTVAL (XEXP (ireg, 1));				\
	      ireg = XEXP (ireg, 0); }					\
	  if (GET_CODE (ireg) == SIGN_EXTEND)				\
	    fprintf (FILE, "pc@(L#%d,%s:w",			\
		     CODE_LABEL_NUMBER (XEXP (addr, 0)),		\
		     reg_names[REGNO (XEXP (ireg, 0))]); 		\
	  else								\
	    fprintf (FILE, "pc@(L#%d(%s:l",			\
		     CODE_LABEL_NUMBER (XEXP (addr, 0)),		\
		     reg_names[REGNO (ireg)]);				\
	  if (scale != 1) fprintf (FILE, "*%d", scale);			\
	  putc (')', FILE);						\
	  break; }							\
      if (ireg != 0 || breg != 0)					\
	{ int scale = 1;						\
	  if (breg == 0)						\
	    abort ();							\
	  fprintf (FILE, "%s@(", reg_names[REGNO (breg)]);		\
	  if (addr != 0)						\
	    output_addr_const (FILE, addr);				\
	  if ((ireg != 0) && (addr != 0))                               \
              putc (',', FILE);						\
	  if (ireg != 0 && GET_CODE (ireg) == MULT)			\
	    { scale = INTVAL (XEXP (ireg, 1));				\
	      ireg = XEXP (ireg, 0); }					\
	  if (ireg != 0 && GET_CODE (ireg) == SIGN_EXTEND)		\
	    fprintf (FILE, "%s:w", reg_names[REGNO (XEXP (ireg, 0))]);	\
	  else if (ireg != 0)						\
	    fprintf (FILE, "%s:l", reg_names[REGNO (ireg)]);		\
	  if (scale != 1) fprintf (FILE, "*%d", scale);			\
	  putc (')', FILE);						\
	  break;							\
	}								\
      else if (reg1 != 0 && GET_CODE (addr) == LABEL_REF)		\
	{ fprintf (FILE, "pc@(L#%d,%s:w)",			\
		   CODE_LABEL_NUMBER (XEXP (addr, 0)),			\
		   reg_names[REGNO (reg1)]);				\
	  break; }							\
    default:								\
      if (GET_CODE (addr) == CONST_INT					\
	  && INTVAL (addr) < 0x8000					\
	  && INTVAL (addr) >= -0x8000)					\
	fprintf (FILE, "%d", INTVAL (addr));				\
      else if (TARGET_32BITDATA && global_data_ref_p (addr))  \
	{  \
	  fprintf (FILE, "a5@(");  \
	  output_addr_const (FILE, addr);			\
	  fprintf (FILE, ")");  \
	}  \
      else  \
        output_addr_const (FILE, addr);				\
    } 

#define PRINT_OPERAND_ADDRESS_MPW(FILE, ADDR)  \
  switch (GET_CODE (addr))						\
    {									\
    case REG:								\
      fprintf (FILE, "(%s)", reg_names[REGNO (addr)]);			\
      break;								\
    case PRE_DEC:							\
      fprintf (FILE, "-(%s)", reg_names[REGNO (XEXP (addr, 0))]);	\
      break;								\
    case POST_INC:							\
      fprintf (FILE, "(%s)+", reg_names[REGNO (XEXP (addr, 0))]);	\
      break;								\
    case PLUS:								\
      reg1 = 0;	reg2 = 0;						\
      ireg = 0;	breg = 0;						\
      offset = 0;							\
      if (CONSTANT_ADDRESS_P (XEXP (addr, 0)))				\
	{								\
	  offset = XEXP (addr, 0);					\
	  addr = XEXP (addr, 1);					\
	}								\
      else if (CONSTANT_ADDRESS_P (XEXP (addr, 1)))			\
	{								\
	  offset = XEXP (addr, 1);					\
	  addr = XEXP (addr, 0);					\
	}								\
      if (GET_CODE (addr) != PLUS) ;					\
      else if (GET_CODE (XEXP (addr, 0)) == SIGN_EXTEND)		\
	{								\
	  reg1 = XEXP (addr, 0);					\
	  addr = XEXP (addr, 1);					\
	}								\
      else if (GET_CODE (XEXP (addr, 1)) == SIGN_EXTEND)		\
	{								\
	  reg1 = XEXP (addr, 1);					\
	  addr = XEXP (addr, 0);					\
	}								\
      else if (GET_CODE (XEXP (addr, 0)) == MULT)			\
	{								\
	  reg1 = XEXP (addr, 0);					\
	  addr = XEXP (addr, 1);					\
	}								\
      else if (GET_CODE (XEXP (addr, 1)) == MULT)			\
	{								\
	  reg1 = XEXP (addr, 1);					\
	  addr = XEXP (addr, 0);					\
	}								\
      else if (GET_CODE (XEXP (addr, 0)) == REG)			\
	{								\
	  reg1 = XEXP (addr, 0);					\
	  addr = XEXP (addr, 1);					\
	}								\
      else if (GET_CODE (XEXP (addr, 1)) == REG)			\
	{								\
	  reg1 = XEXP (addr, 1);					\
	  addr = XEXP (addr, 0);					\
	}								\
      if (GET_CODE (addr) == REG || GET_CODE (addr) == MULT		\
	  || GET_CODE (addr) == SIGN_EXTEND)				\
	{ if (reg1 == 0) reg1 = addr; else reg2 = addr; addr = 0; }	\
      if (offset != 0) { if (addr != 0) abort (); addr = offset; }	\
      if ((reg1 && (GET_CODE (reg1) == SIGN_EXTEND			\
		    || GET_CODE (reg1) == MULT))			\
	  || (reg2 != 0 && REGNO_OK_FOR_BASE_P (REGNO (reg2))))		\
	{ breg = reg2; ireg = reg1; }					\
      else if (reg1 != 0 && REGNO_OK_FOR_BASE_P (REGNO (reg1)))		\
	{ breg = reg1; ireg = reg2; }					\
      if (ireg != 0 && breg == 0 && GET_CODE (addr) == LABEL_REF)	\
        { int scale = 1;						\
	  if (GET_CODE (ireg) == MULT)					\
	    { scale = INTVAL (XEXP (ireg, 1));				\
	      ireg = XEXP (ireg, 0); }					\
	  if (GET_CODE (ireg) == SIGN_EXTEND)				\
	    fprintf (FILE, "L#%d(pc,%s.w",			\
		     CODE_LABEL_NUMBER (XEXP (addr, 0)),		\
		     reg_names[REGNO (XEXP (ireg, 0))]); 		\
	  else								\
	    fprintf (FILE, "L#%d(pc,%s.l",			\
		     CODE_LABEL_NUMBER (XEXP (addr, 0)),		\
		     reg_names[REGNO (ireg)]);				\
	  if (scale != 1) fprintf (FILE, "*%d", scale);			\
	  putc (')', FILE);						\
	  break; }							\
      if (ireg != 0 || breg != 0)					\
	{ int scale = 1;						\
	  if (breg == 0)						\
	    abort ();							\
	  if (addr != 0)						\
	    output_addr_const (FILE, addr);				\
	  fprintf (FILE, "(%s", reg_names[REGNO (breg)]);		\
	  if (ireg != 0)						\
	    putc (',', FILE);						\
	  if (ireg != 0 && GET_CODE (ireg) == MULT)			\
	    { scale = INTVAL (XEXP (ireg, 1));				\
	      ireg = XEXP (ireg, 0); }					\
	  if (ireg != 0 && GET_CODE (ireg) == SIGN_EXTEND)		\
	    fprintf (FILE, "%s.w", reg_names[REGNO (XEXP (ireg, 0))]);	\
	  else if (ireg != 0)						\
	    fprintf (FILE, "%s.l", reg_names[REGNO (ireg)]);		\
	  if (scale != 1) fprintf (FILE, "*%d", scale);			\
	  putc (')', FILE);						\
	  break;							\
	}								\
      else if (reg1 != 0 && GET_CODE (addr) == LABEL_REF)		\
	{ fprintf (FILE, "L#%d(pc,%s.w)",			\
		   CODE_LABEL_NUMBER (XEXP (addr, 0)),			\
		   reg_names[REGNO (reg1)]);				\
	  break; }							\
    default:								\
      if (GET_CODE (addr) == CONST_INT					\
	  && INTVAL (addr) < 0x8000					\
	  && INTVAL (addr) >= -0x8000)					\
	fprintf (FILE, "%d", INTVAL (addr));				\
      else if (TARGET_32BITDATA && global_data_ref_p (addr))  \
	{  \
	  fprintf (FILE, "(");  \
	  output_addr_const (FILE, addr);			\
	  fprintf (FILE, ",a5)");  \
	}  \
      else  \
        output_addr_const (FILE, addr);				\
    }

#define PRINT_OPERAND_ADDRESS(FILE, ADDR)  \
{ register rtx reg1, reg2, breg, ireg;					\
  register rtx addr = ADDR;						\
  rtx offset;								\
  if (TARGET_GAS) PRINT_OPERAND_ADDRESS_GAS(FILE, ADDR) \
  else PRINT_OPERAND_ADDRESS_MPW(FILE, ADDR) }

/* No library stuff, MPW compilers never invoke the linker themselves. */

#define ASM_OUTPUT_ASCII(FILE,STRING,SIZE)  \
  output_mpw_string(FILE, STRING, SIZE);

/* At the end of a function module, prepare for the next module. */

#define ASM_FUNCTION_END(FILE) \
{  extern char *current_function_name; \
   if (!TARGET_GAS) fprintf (FILE, "\tENDP\t; %s\n", current_function_name);  \
}

/* At the end of a variable module, prepare for the next module. */

#define ASM_VARIABLE_END(FILE,NAME) if (!TARGET_GAS) \
{  \
   fprintf (FILE, "\tENDP\t; ");  \
   assemble_name (FILE, NAME);  \
   fprintf (FILE, "\n");  \
}

/* At the end of a variable module, prepare for the next module. */

#define ASM_CONSTANT_END(FILE,NUM,EXP) if (!TARGET_GAS) \
{  \
  if (TARGET_B && TREE_CODE(EXP) == STRING_CST && current_function_decl)  \
    {  \
    }  \
  else  \
    {  \
      fprintf (FILE, "\tENDP\n");  \
    }  \
}

/* Write out a segment name. */

#define ASM_SEGMENT_NAME(FILE,NAME)  \
  if (! TARGET_NOSEG) fprintf(FILE, "\tSEG '%s'\n", NAME);

/* Write out an import of a name not in this module. */

#define ASM_IMPORT_NAME(FILE, NAME)  \
{ \
  (((NAME)[0] == '2') ? data_section () : text_section ());  \
  fprintf (FILE, "\tIMPORT ");  \
  assemble_name (FILE, NAME);  \
  fprintf (file, "\n");  \
}

/* At the end of the whole file, spit out a last directive. */

#define ASM_FILE_END(FILE) \
  if (!TARGET_GAS) fprintf(FILE, "\tEND\n");

/* Magic for the handling of 10/12 byte long doubles. */

#ifdef MPW
#define REAL_VALUE_TYPE extended
#else

/* Don't bother to try to do anything special.  Default will be to use
   'double', which is adequate except for trying to cross-compile extended
   float constants on a machine that doesn't support them, which I don't
   want to bother with right now.  Note that cross-compilation of vars,
   arith, etc, is still ok - only constants might be wrong. */

#endif

/* Machinery to add math-oriented builtin functions. */

#define ADDITIONAL_BUILTIN_SYMS  \
  BUILT_IN_SIN,  \
  BUILT_IN_COS,  \
  BUILT_IN_TAN,  \
  BUILT_IN_ASIN,  \
  BUILT_IN_ACOS,  \
  BUILT_IN_ATAN,  \
  BUILT_IN_SINH,  \
  BUILT_IN_COSH,  \
  BUILT_IN_TANH,  \
  BUILT_IN_EXP,  \
  BUILT_IN_LOG,  \
  BUILT_IN_LOG10,  \
  BUILT_IN_SQRT,

#define ADDITIONAL_BUILTINS  \
  if (TARGET_ELEMS881)  \
    {  \
      builtin_function ("cos", extended_ftype_extended, BUILT_IN_COS);  \
      builtin_function ("sin", extended_ftype_extended, BUILT_IN_SIN);  \
      builtin_function ("tan", extended_ftype_extended, BUILT_IN_TAN);  \
      builtin_function ("asin", extended_ftype_extended, BUILT_IN_ASIN);  \
      builtin_function ("acos", extended_ftype_extended, BUILT_IN_ACOS);  \
      builtin_function ("atan", extended_ftype_extended, BUILT_IN_ATAN);  \
      builtin_function ("sinh", extended_ftype_extended, BUILT_IN_SINH);  \
      builtin_function ("cosh", extended_ftype_extended, BUILT_IN_COSH);  \
      builtin_function ("tanh", extended_ftype_extended, BUILT_IN_TANH);  \
      builtin_function ("exp", extended_ftype_extended, BUILT_IN_EXP);  \
      builtin_function ("log", extended_ftype_extended, BUILT_IN_LOG);  \
      builtin_function ("log10", extended_ftype_extended, BUILT_IN_LOG10);  \
      builtin_function ("sqrt", extended_ftype_extended, BUILT_IN_SQRT);  \
    }

/* A big chunk of code for expression expansion. */
/* Note:  We are passing XFmode rather than mode to expand_unop.  This is
   a hack as mode equals VOIDmode when this is called and it should equal
   XFmode.  This change should be fairly safe as we know that these functions
   use XFmode only.  This worked with all the test cases we could find.  */
#define ADDITIONAL_EXPAND_BUILTIN  \
    case BUILT_IN_SIN:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, sin_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_COS:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, cos_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_TAN:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, tan_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_ASIN:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, asin_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_ACOS:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, acos_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_ATAN:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, atan_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_SINH:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, sinh_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_COSH:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, cosh_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_TANH:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, tanh_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_EXP:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, exp_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_LOG:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, log_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_LOG10:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, log10_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  \
    case BUILT_IN_SQRT:  \
      op0 = expand_expr (TREE_VALUE (arglist), subtarget, VOIDmode, 0);  \
      target = expand_unop (XFmode, sqrt_optab, op0, target, 1);  \
      if (target == 0) abort ();  \
      return target;  
